% this file is published under Creative Commons BY-NC-SA
% 
% Assimilating genome-scale metabolic reconstructions with modelBorgifier
% in preparation
%
% John T. Sauls and Joerg M. Buescher
% BRAIN Aktiengesellschaft
% Microbial Production Technologies Unit
% Quantitative Biology and Sequencing Platform
% Darmstaeter Str. 34-36
% 64673 Zwingenberg, Germany
% www.brain-biotech.de
% jrb@brain-biotech.de
%
%
function nameList = makeNamesUnique(nameList, varargin)
%makeNamesUnique identifies duplicate names in a list of names (cell array)
% and prompts the user to input new names. 
%
%nameList = makeNamesUnique(nameList, [nameInfo])
%
%INPUTS
%nameList   Cell array of entires that contains duplicates.
%
%OPTIONAL INPUTS
%nameInfo   Cell array the same length as nameList that contains
%           information pertaining to the matching entry in nameList. This
%           array can help the renaming process.
%
%OUPUTS
%nameList   Version of nameList with all unique entires.
%
%CALLED BY
% verifyModel

%% Declare variables.
if (nargin > 1)
    nameInfo = varargin{1} ;
end

% List of unique names, sorted.
uniqList = unique(nameList) ;


% Automatically rename entities?
[names,cnt] = countUnique(nameList);
nameInd = find(cnt > 1);
if ~isempty(nameInd)
    
    fprintf('%d non-unique names. ',length(nameInd))
    renameFlag = input(['Rename automatically?\n' ...
                     'Otherwise manual renaming will proceed. (y/N): '],...
                      's') ;
    if strcmpi(renameFlag,'y') || strcmpi(renameFlag,'yes')
        for i = 1:length(nameInd)
            thisName = names{nameInd(i)} ;
            fprintf('Renaming non-unique name %s (%d instances): ',...
                thisName,cnt(nameInd(i)));
            searchString = regexprep(thisName,'\[','\\[') ;
            searchString = regexprep(searchString,'\]','\\]') ;
            searchString = ['^' searchString '$'] ;
            IDs = find(~cellfun(@isempty,regexp(nameList,searchString))) ;
            if isempty(IDs)
                IDs = find(strcmp(nameList,thisName))  ;
            end
            for j = 1:length(IDs)
                if regexp(nameList{IDs(j)},'\[\w\]$')
                    nameList{IDs(j)} = [thisName(1:end-3) '_' num2str(j) ...
                        thisName(end-2:end)] ;
                else
                    nameList{IDs(j)} = [thisName '_' num2str(j)] ;
                end
                fprintf('%s\t',nameList{IDs(j)}) ;
            end
            fprintf('\n')
        end        
    end
end

%% Find duplicate names. 
for iName = 1:length(uniqList) ;
    dupPos = find(strcmp(uniqList{iName},nameList)) ;
    % If there are duplicates.
    if length(dupPos) > 1
        dupNo = length(dupPos) ; 
        % Duplicate name
        dupName = uniqList{iName} ;
        fprintf('%s has duplicates:\n',dupName) 
        % Go through each duplicate and print info.
        for iDup = 1:dupNo
            if exist('nameInfo','var')
                string = [num2str(dupPos(iDup)), '\t', ...
                          nameInfo{dupPos(iDup)}, '\n'] ;
                fprintf(string) 
            else
                fprintf('%d\n',num2str(dupPos(iDup)))
            end
        end
        % Ask for how names should be renamed.
        for iDup = 1:dupNo
            while 1
                prompt = ['Rename ' num2str(dupPos(iDup)) ' as: '] ;
                newName = input(prompt,'s') ;
                alreadytherepos = strcmp(newName,nameList) ;
                alreadytherepos(dupPos) = false ;
                % Check to see if name is already taken
                if sum(alreadytherepos) == 0
                    nameList{dupPos(iDup)} = newName ;
                    break
                else
                    errorPrompt = 'Name already taken. Proceed (y/n)?: ' ;
                    proceed = input(errorPrompt,'s') ;
                    if strcmpi(proceed,'y') || strcmpi(proceed,'yes')
                        nameList{dupPos(iDup)} = newName ;
                        break
                    else
                        continue
                    end
                end
            end
        end
    end
end

%% Check to make sure it worked.
if length(unique(nameList)) ~= length(nameList)
    fprintf('ERROR: Duplicate names still exist!\n') 
end

